library(maptools)
library(rgdal)
library(RColorBrewer)
library(classInt)
library(zipcode)
library(spdep)
library(INLA)
library(sampling)
library(MASS)
library(mgcv)
library(plyr)

## Cond.cCAR function generates the conditional spatial effects 
## when effects of certain areas are given.
## B: the neighborhood matrix
## rho: correlation between neighbors
## tau: precision parameter
## c: index of the regions with fixed effects
## cond.eff: the given fixed effects
## n will be set to 1 as default
Cond.rCAR <- function(B, rho, tau, n, c, cond.eff){
    l <- nrow(B)
    SB <- Matrix(B, sparse = TRUE)
    d <- rowSums(SB)
    SW <- SB/d
    M <- diag(x = d)
    SqrtInvM <- solve(sqrt(M)) 
    MWM <- SqrtInvM %*% B %*% SqrtInvM 
    MinEigen <- min(eigen(MWM, symmetric = TRUE)$values)
    if (rho <= 1/MinEigen || rho > 1) {stop("Invalid rho")}
    Q <- tau^2 * (M - rho * B)
    SIGMA <- solve(Q)
    A <- matrix(0, length(c)+1, l)
    A[1, ] <- 1
    for (i in 2: (length(c)+1)){
        A[i, c[(i-1)]] <- 1
    }
    e <- c(0, cond.eff)
    x <- mvrnorm(n, rep(0, l), solve(Q))
    out <- x - SIGMA %*% t(A) %*% solve(A %*% SIGMA %*% t(A)) %*% (A %*% x - e)
}


## genSP function specifically generate spatial effects for Ohio
## shp.Ohio: the shapefile of Ohio
## mat.Ohio: the neighborhood matrix of Ohio
## cond: the given fixed spatial effects
## position.0: the positions of isolated areas
## rho: correlation between neighbors
## tau: precision parameter
genSP <- function(X, mat.Ohio, zip, cond, position.0, rho, tau){
    ## Extract zipcodes of the largest six cities of Ohio
    columbus <- zip[which(zip >= 43201 & zip <= 43291)]
    cleveland <- zip[which(zip >= 44101 & zip <= 44199)]
    cincinnati <- zip[which(zip >= 45201 & zip <= 45299)]
    toledo <- zip[which(zip >= 43601 & zip <= 43699)]
    akron <- zip[which(zip >= 44301 & zip <= 44399)]
    dayton <- zip[which(zip >=45390 & zip <= 45490)]

## Get positions of these zipcodes in the shapefile
    columbus.index <- as.matrix(match(columbus, X$ZCTA5CE10))
    cincinnati.index <- as.matrix(match(cincinnati, X$ZCTA5CE10))
    cleveland.index <- as.matrix(match(cleveland, X$ZCTA5CE10))
    toledo.index <- as.matrix(match(toledo, X$ZCTA5CE10))
    akron.index <- as.matrix(match(akron, X$ZCTA5CE10))
    dayton.index <- as.matrix(match(dayton, X$ZCTA5CE10))


## Assign positive effects for these zipcodes
## Larger city gets higher values
    columbus.eff <- cbind(columbus.index, rep(cond[1], length(columbus.index)))
    cincinnati.eff <- cbind(cincinnati.index, rep(cond[2], length(cincinnati.index)))
    cleveland.eff <- cbind(cleveland.index, rep(cond[3], length(cleveland.index)))
    toledo.eff <- cbind(toledo.index, rep(cond[4], length(toledo.index)))
    akron.eff <- cbind(akron.index, rep(cond[5], length(akron.index)))
    dayton.eff <- cbind(dayton.index, rep(cond[6], length(dayton.index)))

## Combine these areas and their assigned effects
    cond.eff <- rbind(dayton.eff, columbus.eff, cincinnati.eff, cleveland.eff, toledo.eff, akron.eff)
    colnames(cond.eff) <- c("index", "effect")
    cond.eff <- as.data.frame(cond.eff)

## Due to these isolated areas, we need to adjust positions of other areas
    c <- cond.eff$index
    for(i in 1: length(position.0)){
        c[cond.eff$index > position.0[i]] <- c[cond.eff$index > position.0[i]] - 1 
    }
    out <- c(as.matrix(Cond.rCAR(mat.Ohio[-position.0, -position.0], rho, tau, 1, c, cond.eff$effect)))
    index <- seq(1,length(zip), 1)
    sp.eff <- rep(0, length(zip))
    sim.eff <- as.data.frame(cbind(index, sp.eff))
    sim.eff[-c(position.0), ]$sp.eff <- out
    sim.eff$sp.eff
}

## genData simulates data for all analysis in the paper
## n: total number of observations to simulate
## zip: zipcode list of Ohio
## beta.L: the true fixed effects for logistic part
## beta.N: the true fixed effects for normal part
## weight.zip: percentage of the total population of Ohio each zipcode area takes (Census)
## weight.age: percentage of the total population of Ohio each age bracket takes(Census)
## sigma: random error
## sigma.L: random error of the unstructured effect for logistic part
## sigma.N: random error of the unstructured effect for normal part
## spatial.effect.L: spatial effects generated by genSP for logistic part
## spatial.effect.N: spatial effects generated by genSP for normal part
genData <- function(n, zip, beta.L, beta.N, weight.zip, weight.age, sigma, sigma.L, sigma.N, spatial.effect.L, spatial.effect.N){
  zipcode <- sample(zip, n, replace = TRUE, prob = weight.zip)
  age <- sample(0:17, n, replace = TRUE, prob = weight.age)*5 + sample(0:4, n, replace = TRUE)
  gender <- rbinom(n, 1, 0.5)
  income <- rnorm(n, 0, 1)
  age.eff <- c((seq(0, 25, 1) - 25)^2/200, (seq(26, 89, 1) - 25)^2/600)
  age.eff <- as.data.frame(cbind(seq(0,89,1), age.eff))
  colnames(age.eff) <- c("age.ind", "age.effect")
  random.effect.L <- rnorm(length(zip), 0, sigma.L)
  random.effect.N <- rnorm(length(zip), 0, sigma.N)
  sim.eff <- as.data.frame(cbind(zip, random.effect.L, spatial.effect.L, random.effect.N, spatial.effect.N))
  colnames(sim.eff) <- c("zip", "random.effect.L", "spatial.effect.L", "random.effect.N", "spatial.effect.N")
  X <- cbind(age, gender, income, zipcode)
  X <- merge(X, age.eff, by.x = "age", by.y = "age.ind")
  X <- merge(X, sim.eff, by.x = "zipcode", by.y = "zip")
  X$age.effect <- scale(X$age.effect, center = TRUE, scale = FALSE)
  design.X <- cbind(rep(1, n), X$gender, X$income)
  eta <- design.X %*% beta.L + X$age.effect
  eta1 <- eta + X$spatial.effect.L + X$random.effect.L
  mu_y <- binomial()$linkinv(eta1)
  cost <- rbinom(n, mu_y, size = 1)
  age.eff.n <- scale(X$age.effect[cost==1], center = TRUE, scale = FALSE)
  z <- design.X[cost==1, ] %*% beta.N + age.eff.n + X$spatial.effect.N[cost == 1] + X$random.effect.N[cost == 1] + rnorm(sum(cost), 0, sigma)
  cost[cost == 1] <- exp(z)        
  out <- as.data.frame(cbind(cost, X$age, design.X[, -1], X$zipcode, X$age.effect))
  colnames(out) <- c("cost", "age", "gender", "income", "zipcode", "age.eff")
  out
}


## Read in population data of Ohio
## Weights in the data simulation are determined by proportion of population of each zipcode area.
pop <- read.table("./PopbyZip.csv", sep = ",", header = TRUE)
pop$weight <- pop$population/sum(pop$population)

## Read in population distribution by age
pop.age <- read.table("./PopbyAge.csv", sep = ",", header = TRUE)

ohio_zip_county <- read.table("ohio_zip_county.csv", sep = ",", header = TRUE)
ohio_zip_county$County <- as.character(ohio_zip_county$County)

## Read in zipcode level shape file of Ohio
## Construct neighborhood structure
X <- readShapePoly("./tl_2010_39_zcta510/tl_2010_39_zcta510.shp")
nb.Ohio <- poly2nb(X)
mat.Ohio <- nb2mat(nb.Ohio, style = "B", zero.policy = TRUE)

X.county <- readShapePoly("./tl_2010_39_county10/tl_2010_39_county10.shp")
nb.Ohio.county <- poly2nb(X.county)
mat.Ohio.county <- nb2mat(nb.Ohio.county, style = "B", zero.policy = TRUE)

## Locate areas which area isolated
## We assign 0 to these areas when generating spatial effects
a <- apply(mat.Ohio, 1, sum)
position.0 <- as.numeric(which(a == 0))

## Extract zipcodes of Ohio from shapefile
## zip is the zipcode
## ind is the order of zipcode in the shapefile
zip <- X$ZCTA5CE10
zip <- as.numeric(as.character(zip))
ind <- as.numeric(X$ZCTA5CE10)

##
county <- as.character(X.county$NAME10)



beta.L <- c(2, 0.1, 1)
beta.N <- c(6, 1, 0.4)
cond.L <- seq(0.75, 0.5, -0.05)/2
cond.N <- seq(0.75, 0.5, -0.05)
rho.L <- rho.N <- 0.999
tau.L <- 10
tau.N <- 5
sigma.L <- 0.15
sigma.N <- 0.2
sigma <- 0.1
n <- 20000

set.seed(1111)
sp.eff.L <- genSP(X, mat.Ohio, zip, cond.L, position.0, rho.L, tau.L)
sp.eff.N <- genSP(X, mat.Ohio, zip, cond.N, position.0, rho.N, tau.N)
mydat <- genData(n, zip, beta.L, beta.N, pop$weight, pop.age$percentage, sigma, sigma.L, sigma.N, sp.eff.L, sp.eff.N)
mydat.pred <- mydat
## Construct unstructured variable zipcode.un
mydat$zipcode.un <- mydat$zipcode

mydat <- merge(mydat, ohio_zip_county, by.x = "zipcode", by.y = "zipcode", all.x = TRUE)

## match simulated zipcodes and those in the shapefile
mydat$zip.un <- match(mydat$zipcode.un, X$ZCTA5CE10)
mydat$zip <- match(mydat$zipcode, X$ZCTA5CE10)

mydat$county.str <- match(mydat$County, X.county$NAME10)
mydat$county.un <- mydat$county.str

## Construct indicator variable for logistic regression

mydat.n <- mydat[which(mydat$cost > 0), ]
mydat$zero.ind <- as.numeric(mydat$cost > 0)
## Take log of non-zero cost for normal regression
mydat.n$cost.log <- log(mydat.n$cost)

## Construct spline bases for the non-linear age effect
bs.age.n <- as.data.frame(scale(bs(mydat.n$age, df = 5, degree = 3), center = TRUE, scale = FALSE))
colnames(bs.age.n) <- c("A1", "A2", "A3", "A4", "A5")
mydat.n <- cbind(mydat.n, bs.age.n)
bs.age.l <- as.data.frame(scale(bs(mydat$age, df = 5, degree = 3), center = TRUE, scale = FALSE))
colnames(bs.age.l) <- c("A1", "A2", "A3", "A4", "A5")
mydat <- cbind(mydat, bs.age.l)


write.table(mydat, "mydat.csv", sep = ",", row.names = FALSE, col.names = TRUE)
write.table(mydat.n, "mydat_n.csv", sep = ",", row.names = FALSE, col.names = TRUE)
write.table(mydat.pred, "mydat_pred.csv", sep = ",", row.names = FALSE, col.names = TRUE)
## Alternative way to construct spline bases
#mod.l <- gam(zero.ind ~  s(age, bs = "ps", k = 6), data = mydat, family = binomial())
#mod.n <- gam(cost.log ~  s(age, bs = "ps", k = 6), data = mydat.n)

#bs.age.l <- as.data.frame(model.matrix(mod.l)[, -1])
#bs.age.n <- as.data.frame(model.matrix(mod.n)[, -1])
#colnames(bs.age.l) <- c("A1", "A2", "A3", "A4", "A5")
#colnames(bs.age.n) <- c("A1", "A2", "A3", "A4", "A5")
#mydat <- cbind(mydat, bs.age.l)
#mydat.n <- cbind(mydat.n, bs.age.n)


## Plot simulated structured effects for both logistic and normal parts.
pdf("./sim.pdf", width = 7, height = 3.1)
#png("./sim.png", width = 504, height = 220)
par(mfrow = c(1, 2), mgp = c(0, 0, 0), mar = c(0, 0, 0, 0), ps = 9)
X$sim.eff.L <- sp.eff.L
colors <- brewer.pal(9, "Greys")
brks.sim.l <- classIntervals(X$sim.eff.L, n = 9, style = "quantile")
brks.sim.L <- brks.sim.l$brks
colcode <- findInterval(X$sim.eff.L, brks.sim.L, all.inside = TRUE)
plot(X, col = colors[colcode], axes = F)

for(i in 1:9){
  int.c <- paste("[", signif(brks.sim.L[i], digits = 2), ",", signif(brks.sim.L[i+1], digits = 2), ")", sep = "")
  if(i == 1) c <- int.c
  else c <- c(c, int.c)
}
legend("bottomright", 
       legend = rev(c), 
       title = "Effects",
       fill = names(table(colors[colcode])),
       bty = "n", cex = 0.5)

X$sim.eff.N <- sp.eff.N
colors <- brewer.pal(9, "Greys")
brks.sim.n <- classIntervals(X$sim.eff.N, n = 9, style = "quantile")
brks.sim.N <- brks.sim.n$brks
colcode <- findInterval(X$sim.eff.N, brks.sim.N, all.inside = TRUE)
plot(X, col = colors[colcode], axes = F)

for(i in 1:9){
  int.c <- paste("[", signif(brks.sim.N[i], digits = 2), ",", signif(brks.sim.N[i+1], digits = 2), ")", sep = "")
  if(i == 1) c <- int.c
  else c <- c(c, int.c)
}
legend("bottomright", 
       legend = rev(c), 
       title = "Effects",
       fill = names(table(colors[colcode])),
       bty = "n", cex = 0.5)

dev.off()

## Plot unstructured effect
#X$zip.u <- rep(0, 1197)
#X$zip.u[mod$summary.random$zip.un$ID] <- mod$summary.random$zip.un$mean 
#X$zip.s <- mod$summary.random$zip$mean
#X$zip.t <- X$zip.s + X$zip.u
#colors <- brewer.pal(9, "Greys")
#brks.u <- classIntervals(X$zip.t, n = 9, style = "quantile")
#brks.U <- brks.u$brks
#colcode <- findInterval(X$zip.t, brks.U, all.inside = TRUE)
#plot(X, col = colors[colcode], axes = F)

#for(i in 1:9){
#  int.c <- paste("[", signif(brks.U[i], digits = 2), ",", signif(brks.U[i+1], digits = 2), ")", sep = "")
#  if(i == 1) c <- int.c
#  else c <- c(c, int.c)
#}
#legend("bottomright", 
#       legend = rev(c), 
#       title = "Effects",
#       fill = names(table(colors[colcode])),
#       bty = "n", cex = 0.5)




